import unittest

from autoregex import AutoRegex


class TestAutoRegex(unittest.TestCase):

    def test_get_regex(self):
        def _test_get_regex(line, expected):
            a = AutoRegex()
            r = a.get_regexes([line])[0]
            self.assertEqual(r, expected)

        _test_get_regex("say {word}", "^\W*say\W*(?P<word>.*?\w.*?)\W*$")
        _test_get_regex("{word} is the answer",
                        "^\W*(?P<word>.*?\w.*?)\W*is\W+the\W+answer\W*$")
        _test_get_regex("weather in {city} tomorrow",
                        "^\W*weather\W+in\W*(?P<city>.*?\w.*?)\W*tomorrow\W*$")

    def test_get_entities(self):
        def _test_get_entities(query, lines, expected):
            if not isinstance(lines, list):
                lines = [lines]
            a = AutoRegex()
            a.create_regex(lines)
            for e in a.calc_entities(query):
                self.assertEqual(e, expected)

        _test_get_entities("say hello", "say {word}",
                           {"word": "hello"})
        _test_get_entities("my name is Bond", "my name is {name}",
                           {"name": "Bond"})
        _test_get_entities("I a going to a party tomorrow tonight",
                           "I am going to {place} tomorrow night",
                           {"place": "a party"})
        _test_get_entities("first choice",
                           ["{number} choice", "{number} option"],
                           {"number": "first"})
        _test_get_entities("first option",
                           ["{number} choice", "{number} option"],
                           {"number": "first"})

    def test_get_matches(self):
        def _test_get_matches(query, lines, expected):
            if not isinstance(lines, list):
                lines = [lines]
            a = AutoRegex()
            a.create_regex(lines)
            for e in a.calc_matches(query):
                self.assertEqual(e, expected)

        _test_get_matches("say hello", "say {word}",
                          {'entities': {"word": "hello"},
                           'query': 'say hello',
                           'regex': '^\W*say\W*(?P<word>.*?\w.*?)\W*$'
                           }
                          )
        _test_get_matches("I a going to a party tomorrow tonight",
                          "I am going to {place} tomorrow night",
                          {'entities': {"place": "a party"},
                           'query': 'I a going to a party tomorrow tonight',
                           'regex': '^\W*I\W+am\W+going\W+to\W*(?P<place>.*?\w.*?)\W*tomorrow\W+night\W*$'
                           }
                          )
        _test_get_matches("say blip blop, i am a bot", "say {word}",
                          {'entities': {"word": "blip blop, i am a bot"},
                           'query': 'say blip blop, i am a bot',
                           'regex': '^\W*say\W*(?P<word>.*?\w.*?)\W*$'
                           }
                          )


if __name__ == "__main__":
    unittest.main()



import re


class AutoRegex(object):
    def __init__(self):
        self.regex_lines, self.entity_lines = [], {}
        self.regexes, self.entities = [], {}
        self.must_compile = True
        self.i = 0

    def create_regex(self, lines):
        self.must_compile = True
        self.regex_lines.append(lines)

    def _create_pattern(self, line):
        for pat, rep in (
                # === Preserve Plain Parentheses ===
                (r'\(([^\|)]*)\)', r'{~(\1)~}'),  # (hi) -> {~(hi)~}

                # === Convert to regex literal ===
                (re.escape, None),  # a b:c -> a\ b\:c
                (r' {} '.format, None),  # 'abc' -> ' abc '

                # === Unescape Chars for Convenience ===
                (r'\\ ', r' '),  # "\ " -> " "
                (r'\\{', r'{'),  # \{ -> {
                (r'\\}', r'}'),  # \} -> }
                (r'\\#', r'#'),  # \# -> #

                # === Support Parentheses Expansion ===
                (r'(?<!\\{\\~)\\\(', r'(?:'),  # \( -> (  ignoring  \{\~\(
                (r'\\\)(?!\\~\\})', r')'),  # \) -> )  ignoring  \)\~\}
                (r'\\{\\~\\\(', r'\\('),  # \{\~\( -> \(
                (r'\\\)\\~\\}', r'\\)'),  # \)\~\}  -> \)
                (r'\\\|', r'|'),  # \| -> |

                # === Support Special Symbols ===
                (r'(?<=\s)\\:0(?=\s)', r'\w+'),
                (r'#', r'\d'),
                (r'\d', r'\d'),

                # === Space Word Separations ===
                (r'(?<!\\)(\w)([^\w\s}])', r'\1 \2'),  # a:b -> a :b
                (r'([^\\\w\s{])(\w)', r'\1 \2'),  # a :b -> a : b

                # === Make Symbols Optional ===
                (r'(\\[^\w ])', r'\1?'),

                # === Force 1+ Space Between Words ===
                (r'(?<=\w)(\\\s|\s)+(?=\w)', r'\\W+'),

                # === Force 0+ Space Between Everything Else ===
                (r'\s+', r'\\W*'),
        ):
            if callable(pat):
                line = pat(line)
            else:
                line = re.sub(pat, rep, line)
        return line

    def _create_regex_pattern(self, line):
        line = self._create_pattern(line)
        replacements = {}
        for ent_name in set(re.findall(r'{([a-z_:]+)}', line)):
            replacements[ent_name] = r'(?P<{}>.*?\w.*?)'.format(ent_name)
        for ent_name, ent in self.entities.items():
            ent_regex = r'(?P<{}>{})'
            replacements[ent_name] = ent_regex.format(
                ent_name.replace(':', '__colon__'), ent)
        for key, value in replacements.items():
            line = line.replace('{' + key + '}', value.format(self.i), 1)
            self.i += 1
        return '^{}$'.format(line)

    def _create_regexes(self, lines):
        return [
            re.compile(self._create_regex_pattern(line), re.IGNORECASE)
            for line in sorted(lines, key=len, reverse=True)
            if line.strip()
        ]

    def get_regexes(self, lines):
        regexes = []
        for line in lines:
            regexes.append(self._create_regex_pattern(line))
        return regexes

    def _compile(self):
        self.entities = {
            ent_name: r'({})'.format('|'.join(
                self._create_pattern(line) for line in lines if line.strip()
            ))
            for ent_name, lines in self.entity_lines.items()
        }
        self.regexes = [
            self._create_regexes(lines) for lines in self.regex_lines
        ]
        self.must_compile = False

    def _calc_entities(self, query, regexes):
        for regex in regexes:
            match = regex.match(query)
            if match:
                yield {
                    k.rsplit('__', 1)[0].replace('__colon__', ':'): v.strip()
                    for k, v in match.groupdict().items() if v
                }

    def _calc_matches(self, query, regexes):
        for regex in regexes:
            match = regex.match(query)
            if match:
                yield {
                    k.rsplit('__', 1)[0]: v
                    for k, v in match.groupdict().items() if v
                }

    def calc_entities(self, query):
        if self.must_compile:
            self._compile()
        for regexes in self.regexes:
            entities = list(self._calc_entities(query, regexes))
            if entities:
                yield min(entities, key=lambda x: sum(map(len, x.values())))

    def calc_matches(self, query):
        if self.must_compile:
            self._compile()
        for regexes in self.regexes:
            entities = self._calc_matches(query, regexes)
            regex = str(regexes[0]) \
                .replace("re.compile('", "") \
                .replace("', re.IGNORECASE)", "") \
                .replace("\\\\", "\\")
            for ent in entities:
                yield {
                    'query': query,
                    'entities': ent,
                    'regex': regex
                }
